# -*- coding: utf-8 -*-

"""Module for generating simulated arrays and visilibity data.  The chief 
functions of this module are:

buildSimArray
  given a station object, a list of stands, and a list of frequencies, build 
  a AIPY AntennaArray-like object.  This module can also generate AntennaArray 
  objects with positional errors by setting the 'PosError' keyword to a 
  positive value.

buildSimData
  given a SimArray and a list of aipy.src sources, build up a collection of 
  visibilities for a given set of Julian dates

scaleData
  given a dictionary of simulated visibilities from buildSimData, apply 
  antenna-based gains and delays to the visibilities

shiftData
  given a dictionary of simulated visibilities from buildSimData, shift the uvw 
  coordinates of the visibilities.
  .. note::

	This only changes the uvw values and does not phase-shift the data.

The format of the data dictionaries mentioned above is:

primary keys
  The primary keys store the major aspects of the visiblity data, e.g., 
  frequency coverage, baseline pairs, uvw coordinates, etc.  Valid keys are:
    * *freq* - list of frequencies used in Hz
    * *isMasked* - whether or not the visibility data have been masked 
      (numpy.compress'd)
    * *bls* - list of baselines in (stand 1, stand2) format
    * *uvw* - list of uvw coordinates as 3-element numpy arrays
    * *vis* - list of visibility numpy arrays
    * *wgt* - list of weight arrays with the same length as the visilitity arrays
    * *msk* - list of mask arrays used for the data.  1 = masked, 0 = valid
    * *jd*  - list of Julian dates associated with each list element

secondary keys
  The bls, uvw, vis, wgt, msk, and jd primary keys also have secondary keys that 
  indicate which polarizations are being stored.  Valid keys are:
    * *xx*
    * *yy*
    * *xy*
    * *yx*

In addition to simulation functions, this module includes buildGriddedImage
which takes a dictionary of visibilities and returns and aipy.im.ImgW object.

.. versionchanged:: 0.3
	This module was formerly called lsl.sim.sim
"""

import os
import sys
import aipy
import math
import numpy as n

from lsl import astro
from lsl.common import dp as dp_common
from lsl.common.paths import data as dataPath
from lsl.correlator import uvUtils

__version__ = '0.2'
__revision__ = '$ Revision: 11 $'
__all__ = ['srcs', 'Antenna', 'AntennaArray', 'buildSimArray', 'buildSimData', 'scaleData', 'shiftData', 'buildGriddedImage', '__version__', '__revision__', '__all__']


# A dictionary of bright sources in the sky to use for simulations
srcs = aipy.src.get_catalog(srcs=['Sun', 'Jupiter', 'cas', 'crab', 'cyg', 'her', 'sgr', 'vir'])


class Antenna(aipy.amp.Antenna):
	"""Modification to the aipy.amp.Antenna class to also store the stand ID 
	number in the Antenna.stand attribute.  This also add a getBeamShape 
	attribute that pulls in the old vis.getBeamShape function."""

	def __init__(self, x, y, z, beam, phsoff=[0.,0.], bp_r=n.array([1]), bp_i=n.array([0]), amp=1, pointing=(0.,n.pi/2,0), stand=0, **kwargs):
		"""New init function that include stand ID number support.  From aipy.amp.Antenna:

		  * x,y z = antenna coordinates in equatorial (ns) coordinates
		  * beam = Beam object (implements response() function)
		  * phsoff = polynomial phase vs. frequency.  Phs term that is linear
		    with freq is often called 'delay'.
		  * bp_r = polynomial (in freq) modeling real component of passband
		  * bp_i = polynomial (in freq) modeling imaginary component of passband
		  * amp = overall multiplicative scaling of gain
		  * pointing = antenna pointing (az,alt).  Default is zenith.
		"""

		aipy.phs.Antenna.__init__(self, x,y,z, beam=beam, phsoff=phsoff)
		self.set_pointing(*pointing)
		self.bp_r = bp_r
		self.bp_i = bp_i
		self.amp = amp
		self.stand = stand
		self._update_gain()

	def get_beam_shape(self, pol='x'):
		"""Return a 360 by 90 by nFreqs numpy array showning the beam pattern of a
		particular antenna in the array.  The first two dimensions of the output 
		array contain the azimuth (from 0 to 359 degrees in 1 degree steps) and 
		altitlude (from 0 to 89 degrees in 1 degree steps)."""

		# Build azimuth and altitude arrays.  Be sure to convert to radians
		az = n.zeros((360,90))
		for i in range(360):
			az[i,:] = i*n.pi/180.0
		alt = n.zeros((360,90))
		for i in range(90):
			alt[:,i] = i*n.pi/180.0

		# The beam model is computed in terms of topocentric coordinates, so make that
		# converseion right quick using the aipy.coord module.
		xyz = aipy.coord.azalt2top(n.concatenate([[az],[alt]]))
		
		# I cannot figure out how to do this all at once, so loop through azimuth/
		# altitude pairs
		resp = n.zeros((360,90,len(aa.get_afreqs())))
		for i in range(360):
			for j in range(90):
				resp[i,j,:] = n.squeeze( self.bm_response(n.squeeze(xyz[:,i,j]), pol=pol) )

		return resp


class AntennaArray(aipy.amp.AntennaArray):
	"""Modification to the aipy.ant.AntennaArray class to add a fuction to 
	retrieve the stands stored in the AntennaArray.ants attribute.  Also add 
	a function to set the array time from a UNIX timestamp."""

	def get_stands(self):
		"""Return a numpy array listing the stands found in the AntennaArray 
		object."""

		stands = []
		for ant in self.ants:
			stands.append(ant.stand)
		
		return numpy.array(stands)

	def set_unixtime(self, timestamp):
		"""Set the array time using a UNIX timestamp (epoch 1970)."""

		self.set_jultime(astro.unix_to_utcjd(time))


def buildSimArray(station, stands, freq, jd=None, PosError=0.0, ForceFlat=False, verbose=False):
	"""Build a AIPY AntennaArray for simulation purposes.  Inputs are a station 
	object defined from the lwa_common module, a numpy array of stand numbers, 
	and a numpy array of frequencies in either Hz of GHz.  Optional inputs are
	a Julian Date to set the array to and a positional error terms that perturbs
	each of the stands in x, y, and z.  The output of this module is an AIPY
	AntennaArray object.

	The shape of the antenna response is either flat (gain of 1 in all 
	directions) or modeled by a collection of spherical harmonics that are poly-
	nomials in frequency.  The spherical harmonics are used if the file 
	'beam_shape.npz' is found in the current directory."""

	# If the frequencies are in Hz, we need to convert to GHz
	freqs = freq.copy()
	if freqs.min() > 1e6:
		freqs /= 1.0e9
	
	# If the beam Alm coefficient file is present, build a more relatistc beam 
	# response.  Otherwise, assume a flat beam
	if not ForceFlat and os.path.exists(os.path.join(dataPath, 'beam-shape.npz')):
		dd = n.load(os.path.join(dataPath, 'beam-shape.npz'))
		coeffs = dd['coeffs']

		deg = coeffs.shape[0]-1
		lmax = int((math.sqrt(1+8*coeffs.shape[1])-3)/2)
		beamShapeDict = {}
		for i in range(deg+1):
			beamShapeDict[i] = n.squeeze(coeffs[-1-i,:])

		if verbose:
			print "Using Alm beam model with %i-order freq. polynomial and %i-order sph. harmonics" % (deg, lmax)
		beam = aipy.amp.BeamAlm(freqs, lmax=lmax, mmax=lmax, deg=deg, nside=128, coeffs=beamShapeDict)
	else:
		if verbose:
			print "Using flat beam model"
		beam = aipy.amp.Beam(freqs)

	if PosError != 0:
		print "WARNING:  Creating array with positional errors between %.3f and %.3f m" % (-PosError, PosError)

	# Build an array of AIPY Antenna objects
	ants = []
	posCache = uvUtils.PositionCache()
	for stand in stands:
		top = posCache.getXYZ(stand)
		top += (2*PosError*n.random.rand(3)-PosError)	# apply a random positional error if needed
		top.shape = (3,)
		eq = n.dot( aipy.coord.top2eq_m(0.0, station.lat), top )
		eq *= 100	# m -> cm
		eq /= aipy.const.c	# cm -> s
		eq *= 1e9	# s -> ns

		delayCoeff = n.zeros(2)

		amp = 1.0 / uvUtils.cableAttenuation(stand)
		
		ants.append( Antenna(eq[0], eq[1], eq[2], beam, phsoff=delayCoeff, amp=amp, stand=stand) )

	# Combine the array of antennas with the array's location to generate an
	# AIPY AntennaArray object
	simAA = AntennaArray(ants=ants, location=station.getAIPYLocation())

	# Set the Julian Data for the AntennaArray object if it is provided.  The try...except
	# clause is used to deal with people who may want to pass an array of JDs in rather than
	# just one.
	if jd is not None:
		try:
			junk = len(jd)
		except:
			simAA.set_jultime(jd)
		else:
			simAA.set_jultime(jd[0])

	return simAA


def __buildSimData(aa, srcs, jd=None, phaseCenter='z', baselines=None, mask=None, verbose=False, count=None, max=None):
	"""Helper function for buildSimData so that buildSimData can be called with 
	a list of Julian Dates and reconstruct the data appropriately."""

	nFreq = (n.squeeze((aa.get_afreqs()))).shape[0]

	# Update the JD if necessary
	if jd is not None:
		if count is not None and max is not None:
			print "Setting Julian Date to %.5f (%i of %i)" % (jd, count, max)
		else:
			print "Setting Julian Date to %.5f" % jd
		aa.set_jultime(jd)
	else:
		jd = aa.get_jultime()

	# Compute the source parameters
	srcs_eq = []
	srcs_jy = []
	srcs_fq = []
	srcs_sh = []
	if verbose:
		print "Sources Used for Simulation:"
	for name,src in srcs.iteritems():
		## Update the source's coordinates
		src.compute(aa)

		## RA/dec -> equitorial 
		srcEQ = src.get_crds(crdsys='eq', ncrd=3)
		srcEQ.shape = (3,1)
		srcs_eq.append( srcEQ )

		## Source flux over the bandpass
		jys = src.get_jys()
		jys.shape = (1,nFreq)
		srcs_jy.append( jys )

		## Frequencies that the fluxes correspond to
		frq = aa.get_afreqs()
		frq.shape = (1,nFreq)
		srcs_fq.append( frq )

		## Source shape parameters
		shp = n.array(src.srcshape)
		shp.shape = (3,1)
		srcs_sh.append( shp )

		if verbose:
			srcAzAlt = aipy.coord.top2azalt(src.get_crds(crdsys='top', ncrd=3))
			if srcAzAlt[1] < 0:
				print "  %s: below horizon" % name
			else:
				print "  %s: %+.2f altitude; flux mean %.2f kJy" % (name, srcAzAlt[1]*180.0/math.pi, jys.mean()/1e3)

	# Build the simulation cache
	aa.sim_cache( n.concatenate(srcs_eq, axis=1), 
				jys=n.concatenate(srcs_jy, axis=0),
				mfreqs=n.concatenate(srcs_fq, axis=0),
				srcshapes=n.concatenate(srcs_sh, axis=1) )

	# Build the simulated data.  If no baseline list is provided, build all 
	# baselines avaliable
	if baselines is None:
		baselines = uvUtils.getBaselines(n.zeros(len(aa.ants)), Indicies=True)

	# Define output data structure
	UVData = {'freq': n.squeeze(aa.get_afreqs()*1e9), 'uvw': {}, 'vis': {}, 'wgt': {}, 'msk': {}, 'bls': {}, 'jd': {}}
	if mask is None:
		UVData['isMasked'] = False
	else:
		UVData['isMasked'] = True

	# Go!
	if phaseCenter is not 'z':
		phaseCenter.compute(aa)

	pols = ['xx', 'yy', 'xy', 'yx']
	for pol in pols:
		UVData['bls'][pol] = []
		UVData['uvw'][pol] = []
		UVData['vis'][pol] = []
		UVData['wgt'][pol] = []
		UVData['msk'][pol] = []
		UVData['jd'][pol] = []

		if mask is None:
			for (i,j) in baselines:
				if verbose:
					print "Simulating data for baseline %i-%i, pol. %s" % (i,j,pol)

				try:
					crd = aa.gen_uvw(i, j, src=phaseCenter)
					d = aa.sim(i, j, pol=pol)
					d = aa.phs2src(d, phaseCenter, i, j)
				except aipy.phs.PointingError:
					continue

				UVData['bls'][pol].append( (i,j) )
				UVData['uvw'][pol].append( n.squeeze(crd) )
				UVData['vis'][pol].append( d )
				UVData['wgt'][pol].append( n.ones_like(d)*len(UVData['vis'][pol][-1]) )
				UVData['msk'][pol].append( n.zeros_like(UVData['vis'][pol][-1]) )
				UVData['jd'][pol].append( jd )
		else:
			for (i,j),msk in zip(baselines,mask):
				if verbose:
					print "Simulating data for baseline %i-%i, pol. %s" % (i,j,pol)
				
				try:
					crd = aa.gen_uvw(i, j, src=phaseCenter)
					d = aa.sim(i, j, pol=pol)
					d = aa.phs2src(d, phaseCenter, i, j)
				except aipy.phs.PointingError:
					continue

				uvw = aa.gen_uvw(i, j, src=phaseCenter)
				uvw = n.squeeze(crd.compress(n.logical_not(msk), axis=2))
				vis = d.compress(n.logical_not(msk))
				wgt = n.ones_like(vis) * len(vis)

				UVData['bls'][pol].append( (i,j) )
				UVData['uvw'][pol].append( uvw )
				UVData['vis'][pol].append( vis )
				UVData['wgt'][pol].append( wgt )
				UVData['msk'][pol].append( msk )
				UVData['jd'][pol].append( jd )

	return UVData


def buildSimData(aa, srcs, jd=None, phaseCenter='z', baselines=None, mask=None, verbose=False):
	"""Given an AIPY AntennaArray object and a dictionary of sources from 
	aipy.src.get_catalog, returned a data dictionary of simulated data taken 
	at zenith.  Optinally, the data can be masked using some referenced 
	(observed) data set or only a specific sub-set of baselines."""

	nFreq = (aa.get_afreqs()).shape[0]

	# Update the JD if necessary
	if jd is not None:
		try:
			junk = len(jd)
		except TypeError:
			jd = [jd]
	else:
			jd = [aa.get_jultime()]

	# Build up output dictionary
	UVData = {'freq': n.squeeze(aa.get_afreqs()*1e9), 'uvw': {}, 'vis': {}, 'wgt': {}, 'msk': {}, 'bls': {}, 'jd': {}}
	if mask is None:
		UVData['isMasked'] = False
	else:
		UVData['isMasked'] = True
	
	# Loop over polarizations to populate the simulated data set
	pols = ['xx', 'yy', 'xy', 'yx']
	for pol in pols:
		UVData['bls'][pol] = []
		UVData['uvw'][pol] = []
		UVData['vis'][pol] = []
		UVData['wgt'][pol] = []
		UVData['msk'][pol] = []
		UVData['jd'][pol] = []

	# Loop over Julian days to fill in the simulated data set
	jdCounter = 1
	for juldate in jd:
		oBlk = __buildSimData(aa, srcs, jd=juldate, phaseCenter=phaseCenter, baselines=baselines, mask=mask, verbose=verbose, count=jdCounter, max=len(jd))
		jdCounter = jdCounter + 1

		for pol in oBlk['bls']:
			for b,u,v,w,m,j in zip(oBlk['bls'][pol], oBlk['uvw'][pol], oBlk['vis'][pol], oBlk['wgt'][pol], oBlk['msk'][pol], oBlk['jd'][pol]):
				UVData['bls'][pol].append( b )
				UVData['uvw'][pol].append( u )
				UVData['vis'][pol].append( v )
				UVData['wgt'][pol].append( w )
				UVData['msk'][pol].append( m )
				UVData['jd'][pol].append( j )

	return UVData


def buildSimSignals(aa, stands, srcs, length=1024000, jd=None, phaseCenter='z', pol='x', SampleRate=dp_common.fS, IQ=False, verbose=True):
	"""Build up a collection of time series data that represents the output
	voltage of each stand as a function of time for a given collection of
	sources.  Note:  This only seems to work for single sources."""

	eta = 119.9169832*n.pi	# impedance of free space (Ohm)

	Nstand = len(aa.ants)
	freq = n.squeeze((aa.get_afreqs())) * 1e9 # GHz to Hz
	Nfreq = freq.shape[0]
	
	# Compute the source parameters
	srcs_tp = []
	srcs_mp = []
	srcs_jy = []
	srcs_fq = []
	if verbose:
		print "Sources Used for Simulation:"
	for name,src in srcs.iteritems():
		## Update the source's coordinates
		src.compute(aa)

		## Get parameters
		top = src.get_crds(crdsys='top', ncrd=3)	# topo. coords.
		map = src.map							# equitorial -> topo. rotation matrix
		jys = src.get_jys()						# F_nu
		frq = aa.get_afreqs()					# nu

		## Fix the lowest frequencies to avoid problems with the flux blowing up
		## at nu = 0 Hz by replacing flux values below 1 MHz with the flux at 
		## 1 MHz
		Jyat1MHz = jys[ n.where( n.abs(frq-0.001) == n.abs(frq-0.001).min() ) ]
		jys = n.where( frq >= 0.001, jys, Jyat1MHz )

		## Filter out sources that are below the horizon or have no flux
		srcAzAlt = aipy.coord.top2azalt(top) * 180/math.pi
		if verbose:
			if srcAzAlt[1] < 0:
				print "  %s: below horizon" % name
			else:
				print "  %s: %+.2f altitude; flux mean %.2f kJy" % (name, srcAzAlt[1], jys.mean()/1e3)
		if srcAzAlt[1] <= 0 or jys.sum() <= 0:
			continue

		## Save values into the source arrays
		srcs_tp.append( top )
		srcs_mp.append( map )
		srcs_jy.append( jys )
		srcs_fq.append( frq )

	# Get the topocentric coorindates for the zenith
	zen = aipy.coord.azalt2top(n.array([[n.pi/4],[n.pi/2]]))
	zen = n.squeeze(zen)

	# Define the stand position and cable/signal delay caches to the simulator 
	# move along faster
	dlyCache = uvUtils.SignalCache(freq, applyDispersion=True)

	# Update the phase center if necessary
	if phaseCenter == 'z':
		phaseCenterMap = aipy.coord.eq2top_m(0.0, aa.lat)
	else:
		phaseCenter.compute(aa)
		phaseCenterMap = phaseCenter.map

	# Generate the time series response of each signal at each frequency
	t = n.arange(length)/SampleRate * 1e9 # s -> ns
	tdSignals = n.zeros((Nstand, length), dtype=n.complex64)
	temp = n.zeros((len(srcs_tp), Nstand, length), dtype=n.complex64)
	for sc,tp,mp,jy,fq in zip(range(temp.shape[0]), srcs_tp, srcs_mp, srcs_jy, srcs_fq):
		for i,ant in zip(range(Nstand), aa.ants):
			# Zeroth, get the beam response in the direction of the current source for all frequencies
			antResponse = n.squeeze( ant.bm_response(tp, pol=pol) )

			# First, do the geometric delay
			geoDelay = ( n.dot(mp, ant.pos).transpose() )[2] 
			geoDelayPC = ( n.dot(phaseCenterMap, ant.pos).transpose() )[2]
			print geoDelay, geoDelayPC

			# Second, do the cable delay
			Delayat1MHz = dlyCache.cableDelay(stands[i], freq=1.0e6)[0] * 1e9 # s -> ns
			cblDelay = dlyCache.cableDelay(stands[i]) * 1e9 # s -> ns
			# NB: Replace the cable delays below 1 MHz with the 1 MHz value to keep the 
			# delays from blowing up for small f
			cblDelay = n.where( fq >= 0.001, cblDelay, Delayat1MHz )

			for a,j,f,d in zip(antResponse, jy, fq, cblDelay):
				factor = a * n.sqrt(j)
				angle = 2*n.pi*f*(t + (d - (geoDelay-geoDelayPC)))
				temp[sc,i,:] += factor*(n.cos(angle) + 1j*n.sin(angle))

	print temp.shape
	tdSignals = temp.sum(axis=0)
	if not IQ:
		return tdSignals.real
	else:
		return tdSginals


def scaleData(dataDict, amps, delays):
	"""Apply a set of antenna-based gains and phase delays to a data dictionary.
	Returned the new scaled and delayed dictionary."""

	import copy

	# Build the data dictionary to hold the scaled and delayed data
	sclUVData = {'freq': (dataDict['freq']).copy(), 'uvw': {}, 'vis': {}, 'wgt': {}, 'msk': {}, 'bls': {}, 'jd': {}}
	if dataDict['isMasked']:
		sclUVData['isMasked'] = True
	else:
		sclUVData['isMasked'] = False

	# Apply the scales and delays for all polarization pairs found in the original data
	for pol in dataDict['vis'].keys():
		sclUVData['bls'][pol] = copy.copy(dataDict['bls'][pol])
		sclUVData['uvw'][pol] = copy.copy(dataDict['uvw'][pol])
		sclUVData['vis'][pol] = []
		sclUVData['wgt'][pol] = copy.copy(dataDict['wgt'][pol])
		sclUVData['msk'][pol] = copy.copy(dataDict['msk'][pol])
		sclUVData['jd'][pol] = copy.copy(dataDict['jd'][pol])

		for (i,j),vis in zip(dataDict['bls'][pol], dataDict['vis'][pol]):
			sclUVData['vis'][pol].append( vis*amps[i]*amps[j]*n.exp(-1j*2.0*math.pi*(delays[j]-delays[i])) )

	return sclUVData
	

def shiftData(dataDict, aa):
	"""Shift the uvw coordinates in one data dictionary to a new set of uvw 
	coordinates that correspond to a new AntennaArray object.  This is useful
	for looking at how positional errors in the array affect the data."""

	import copy
	
	# Build the data dictionary to hold the shifted data
	sftUVData = {'freq': (dataDict['freq']).copy(), 'uvw': {}, 'vis': {}, 'wgt': {}, 'msk': {}, 'bls': {}, 'jd': {}}
	if dataDict['isMasked']:
		sftUVData['isMasked'] = True
	else:
		sftUVData['isMasked'] = False

	# Apply the coordinate shift all polarization pairs found in the original data
	for pol in dataDict['vis'].keys():
		sftUVData['bls'][pol] = copy.copy(dataDict['bls'][pol])
		sftUVData['uvw'][pol] = []
		sftUVData['vis'][pol] = copy.copy(dataDict['vis'][pol])
		sftUVData['wgt'][pol] = copy.copy(dataDict['wgt'][pol])
		sftUVData['msk'][pol] = copy.copy(dataDict['msk'][pol])
		sftUVData['jd'][pol] = copy.copy(dataDict['jd'][pol])

		for (i,j),m in zip(dataDict['bls'][pol], dataDict['msk'][pol]):
			crds = aa.gen_uvw(i, j, src='z')
			if dataDict['isMasked']:
				crds = crds.compress(n.logical_not(n.logical_not(m), axis=2))
			crds = n.squeeze(crds)
			sftUVData['uvw'][pol].append( crds )

	return sftUVData


def buildGriddedImage(dataDict, MapSize=30, MapRes=0.50, MapWRes=0.10, pol='xx', chan=None):
	"""Given a data dictionary, build an aipy.img.ImgW object of gridded uv data 
	which can be used for imaging.  The ImgW object itself is returned by this 
	function to make it more versatile."""

	im = aipy.img.ImgW(size=MapSize, res=MapRes, wres=MapWRes)

	if chan is not None:
		# Make sure that `chan' is an array by trying to find its length
		try:
			junk = len(chan)
		except TypeError:
			chan = [chan]

		# Build up the data using only the specified channels
		uvw = []
		vis = []
		wgt = []
		for (i,j),d,u,w,m in zip(dataDict['bls'][pol], dataDict['vis'][pol], dataDict['uvw'][pol], dataDict['wgt'][pol], dataDict['msk'][pol]):
			u = u[:,chan]
			u.shape = (3, len(chan))

			uvw.append(u)
			vis.append(n.array([d[chan]]))
			wgt.append(n.array([w[chan]]))
	else:
		uvw = dataDict['uvw'][pol]
		vis = dataDict['vis'][pol]
		wgt = dataDict['wgt'][pol]

	uvw = n.concatenate(uvw, axis=1)
	vis = n.concatenate(vis)
	wgt = n.concatenate(wgt)

	uvw, vis, wgt = im.append_hermitian(uvw, vis, wgts=wgt)
	im.put(uvw, vis, wgts=wgt)

	return im



# -*- coding: utf-8 -*-

"""Module to write VDIF frames.  The implementation of this module is similar
to that of lsl.sim.tbw in that the primary element defined in this module is
a Frame object which as attribute functions that can create a numpy 
representation of the raw frame and write that raw frame to an open file-
handle.

.. seealso::
	:mod:`lsl.sim.tbw`

"""

import math
import numpy
import ephem
import struct

from lsl.common import stations as lwa_common
from lsl.common import dp as dp_common
import lsl.astro as astro

__version__ = '0.1'
__revision__ = '$ Revision: 4 $'
__all__ = ['Frame', '__version__', '__revision__', '__all__']

vdifEpoch = ephem.Date('2000/01/01 00:00:00.00')
unixEpoch = ephem.Date('1970/01/01 00:00:00.00')


class Frame(object):
	"""Object to create and write a VDIF (VLBI Data Interchange Format) 
	frame (version 1, June 26, 2009)."""
	
	def __init__(self, stand=0, time=0, bits=16, data=None, sampleRate=dp_common.fS):
		self.stand = stand
		self.time = time
		self.bits = bits
		self.data = numpy.squeeze(data)
		self.sampleRate = sampleRate
		if data.dtype.kind == 'c':
			self.dataReal = False
		else:
			self.dataReal = True

		# Convert the time from UNIX to epoch and make the data ready to 
		# be written to the disk.
		self.__dataAdjust()
		self.__setEpoch()
		
	def __dataAdjust(self):
		"""Adjust the data to the number of bits required for the frame 
		and convert to unsigned 32-bit integers.  Also, interlace complex
		arrays."""
		
		if self.dataReal:
			interlaced = self.data
		else:
			interlaced = numpy.zeros(2*len(self.data))
			interlaced[0::2] = self.data.real
			interlaced[1::2] = self.data.imag
		
		biased = interlaced + 2**(self.bits-1)
		biased = biased.astype(numpy.uint32) & (2**self.bits-1)
		
		self.data = biased
		
	def __setEpoch(self):
		"""Convert a UNIX timestap in seconds since 1970 to seconds since
		January 1st, 2000 using ephem.Data objects.  Also, calculate the 
		frame number since the beginning of the second."""
		
		# UNIX time to seconds since DJD  = 0
		curEpoch = float(unixEpoch)*astro.SECS_IN_DAY + self.time
		self.epoch = 0
		# Seconds since the VDIF epoch
		epochSeconds = curEpoch - float(vdifEpoch)*astro.SECS_IN_DAY
		# Integer seconds
		self.seconds = long(epochSeconds)
		
		# Compute the frames since the beginning of the second
		frame = (epochSeconds - self.seconds) * (self.sampleRate/len(self.data))
		if self.dataReal:
			self.frame = long(frame)
		else:
			self.frame = long(frame * 2)

	def createRawFrame(self):
		"""Using the data and information stored in the object, create a
		numpy array on unsigned 1-byte integers that represents the frame."""
		
		# Find out how many samples can be packed into a work and build
		# the corresponding numpy array.
		samplesPerWord = int(32 / self.bits)
		raw = numpy.zeros(32 + 4*len(self.data)/samplesPerWord, dtype=numpy.uint8)

		# Valid data, standard (not legacy) 32-bit header, and seconds since 
		# the 01/01/2000 epoch.
		raw[3] = (0 << 7) | (0 << 6) | ((self.seconds >> 24) & 63)
		raw[2] = (self.seconds >> 16) & 255
		raw[1] = (self.seconds >> 8) & 255
		raw[0] = self.seconds & 255

		# Refernce epoch (0 == 01/01/2000) and frame count
		raw[7] = self.epoch & 63
		raw[6] = (self.frame >> 16) & 255
		raw[5] = (self.frame >> 8) & 255
		raw[4] = self.frame & 255

		# VDIF version number, number of channels (just 1), and data frame 
		# length in units to 8-bytes (8 raw array elements)
		raw[11] = (1 << 6) | (0 & 31)
		raw[10] = ((len(raw) / 8) >> 16) & 255
		raw[9] = ((len(raw) / 8) >> 8) & 255
		raw[8] = (len(raw) / 8) & 255

		# Data type, bits per sample, thread ID, and station ID
		# NB:  The thread ID is fixed at `0'
		if self.dataReal:
			raw[15] = (0 << 7) | ((self.bits & 31) << 2) | ((0 >> 8) & 3)
		else:
			raw[15] = (1 << 7) | ((self.bits & 31) << 2) | ((0 >> 8) & 3)
		raw[14] = 0 & 255
		raw[13] = (self.stand >> 8) & 255
		raw[12] = self.stand & 255
		
		# User-defined words 4 through 7 (array entries 16 to 31)
		raw[16:19] = 0
		raw[20:23] = 0
		raw[24:27] = 0
		raw[28:31] = 0
		
		# Data values
		for f in range(32,len(raw),4):
			i = (f - 32) / 4
			word = 0
			for p in range(samplesPerWord):
				word |= (self.data[i*samplesPerWord + p] << self.bits*p)
			raw[f+3] = (word >> 24) & 255
			raw[f+2] = (word >> 16) & 255
			raw[f+1] = (word >> 8) & 255
			raw[f+0] = word & 255
				
		return raw

	def writeRawFrame(self, fh):
		"""Create a numpy representation of the VDIF frame and then write 
		it to the specified file handle."""
		
		rawFrame = self.createRawFrame()
		rawFrame.tofile(fh)



# -*- coding: utf-8 -*-

"""lsl.misc - Miscellanous modules including:
 * mathutil   - math utilities that were part of the lwa_user package, 
 * geodesy    - module for retrieving earth orientation parameters, and 
 * beamformer - post data aquisition beam former for TBW and TBN data."""

import mathutil
import geodesy
import beamformer

# Maybe some day
#import difxconfig



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Python script to read in a S60 file and average it in time.  The output is a 
npz file of the time-averaged spectra and a PNG of the bandpass/waterfall diagram."""

import os
import sys
import numpy
import getopt

from lsl.common.paths import data as dataPath
from lsl.reader import s60

import matplotlib.pyplot as plt


def usage(exitCode=None):
	print """readS60.py - Read in raw S60 files and create a collection of 
time- averaged spectra that are saved to NPZ files.

Usage: readS60.py [OPTIONS] file

Options:
-h --help                  Display this help information
-e --enable-model          Use the CFTOOL bandpass model if 
                            it is present in the current 
                            directory
-q --quiet                 Run readS60 in silent mode
-l --fft-length            Set FFT length (default = 4096)
-t --avg-time              Window to average spectra in time
"""

	if exitCode is not None:
		sys.exit(exitCode)
	else:
		return True


def parseOptions(args):
	config = {}
	# Command line flags - default values
	config['avgTime'] = 10
	config['LFFT'] = 4096
	config['verbose'] = True
	config['enableModel'] = False
	config['args'] = []

	# Read in and process the command line flags
	try:
		opts, args = getopt.getopt(args, "heql:t:", ["help", "enable-model", "quiet", "fft-length=", "avg-time="])
	except getopt.GetoptError, err:
		# Print help information and exit:
		print str(err) # will print something like "option -a not recognized"
		usage(exitCode=2)
	
	# Work through opts
	for opt, value in opts:
		if opt in ('-h', '--help'):
			usage(exitCode=0)
		elif opt in ('-e', '--enable-model'):
			config['enableModel'] = False
		elif opt in ('-q', '--quiet'):
			config['verbose'] = False
		elif opt in ('-l', '--fft-length'):
			config['LFFT'] = int(value)
		elif opt in ('-t', '--avg-time'):
			config['avgTime'] = int(value)
		else:
			assert False
	
	# Add in arguments
	config['args'] = args

	# Return configuration
	return config


def main(args):
	# Parse command line options
	config = parseOptions(args)

	# Length of the FFT
	LFFT = config['LFFT']

	# Primary integration time in seconds.  This sets how large of chunks to read in 
	# from the file at one time.  This is limited by the ammount of memory avaliable 
	# on the host machine and the upper limit on fornax is around 20-30 seconds.  The
	# integration time should be a multiple of two to ensure that the number of 
	# samples read in divisable by 4096.
	integrationTime = config['avgTime']

	for filename in config['args']:
		# The total number of samples in a file is its size divided by 2 bytes per sample
		nSamples = os.path.getsize(filename) / 2
		# The number of seconds in a file is its size divided by the number of bytes per
		# packet, divided by the number of packets per second
		nSeconds = os.path.getsize(filename) / 2 / s60.SampleRate
		print "Working on file: '%s'" % filename
		print "Samples in file: %i" % nSamples
		print "Seconds in file: %i" % nSeconds

		# Open file for reading in binary mode
		fh = open(filename, "rb", buffering=7340000)

		# Frequency array for LFFT channels and a 3.75 MHz sample rate
		w = numpy.arange(-1.875, 1.875, 3.50/LFFT)

		# Work on the data in ``integrationTime" second segments
		timeBlocks = numpy.zeros((nSeconds/integrationTime, LFFT))
		for i in range(timeBlocks.shape[0]):
			if config['verbose']:
				print "Working on time: %i to %i... %6.2f%%" % (i*integrationTime, (i+1)*integrationTime, (100.0*i/timeBlocks.shape[0]))
			data = s60.readChunk(fh, Chunk=integrationTime*5109*734)
			# Work on computing an integrated spectrum.  This is done by reading in a chunk 
			# of length LFFT, computing is power spectrum, and adding it to all of the pre-
			# vious samples.  The min, mean, and max powers are printed out at every 1,000-
			# th step
			avgCount = 0
			integratedPS = numpy.zeros(LFFT)
			for j in range(data.shape[0]/LFFT):
				avgCount = avgCount + 1

				oneChunk = data[j*LFFT:(j+1)*LFFT]
				oneChunk = oneChunk - oneChunk.mean()
				chunkFFT = numpy.fft.fftshift(numpy.fft.fft(oneChunk))
				chunkPS = numpy.abs(chunkFFT)**2.0
				
				# This middle point of the power spectrum is always zero and it
				# always causes problems with the fitting later on.  So, replace
				# it with the average of its neighbors.
				chunkPS[LFFT/2] = 0.5*chunkPS[LFFT/2-1] + 0.5*chunkPS[LFFT/2+1]

				integratedPS = integratedPS + chunkPS
				
				if j % 1000 == 0 and config['verbose']:
					print " %6.2f%%" % (100.0*j*LFFT/data.shape[0])

			# Print out the last integrated power spectrum to the master array
			timeBlocks[i,:] = 10.0*numpy.log10(integratedPS/avgCount)
		fh.close()

		# Save the data so we don't have to go through all that again
		basename = filename[filename.rfind('/')+1:]
		output = basename.replace('.txt', '.npz')
		output = output.replace('.dat', '.npz')
		numpy.savez(output, w=w, timeBlocks=timeBlocks, integrationTime=integrationTime)

		# Bandpass fitting can either be done with Joe's model from the wiki (if in the
		# current directory) or with a 17-th order polynomial.  In general, I found that
		# the higher order works better on some of the bandpasses.
		if config['enableModel']:
			bandpass = s60.getBandpassModel()
			bandpass = bandpass[300:3700]
			bandpass = bandpass - bandpass.mean() + timeBlocks[:,300:3700].mean()
		else:
			bandpassCoeff = numpy.polyfit(w[300:3700], numpy.squeeze(timeBlocks[:,300:3700].mean(axis=0)), 
									17)
			print "Bandpass Polynomial Coefficients: ", bandpassCoeff
			bandpass = numpy.polyval(bandpassCoeff, w[300:3700])

		# Apply the bandpass correction
		for i in range(timeBlocks.shape[0]):
			timeBlocks[i,300:3700] = timeBlocks[i,300:3700] - bandpass

		# Display and write the results to a PNG
		fig = plt.figure()
		ax1 = fig.add_subplot(211)
		ax1.plot(w[300:3700], numpy.squeeze(timeBlocks[:,300:3700]).mean(axis=0)+bandpass, color='green', 
				linestyle=' ', marker='x', markersize=1.0)
		ax1.plot(w[300:3700], bandpass, color='blue', linestyle='-')
		ax1.set_xlabel('Frequncy Bin')
		ax1.set_ylabel('Bandpass')
		ax1.set_xlim([w[300], w[3700]])
		ax1.axis('auto')
		ax2 = fig.add_subplot(212)
		ax2.imshow(timeBlocks[:,300:3700], origin='lower', vmin=0, vmax=0.3, 
					extent=(w[300], w[3700], 0, (timeBlocks.shape[0]+1)*integrationTime))
		ax2.set_xlabel('Frequncy Bin')
		ax2.set_ylabel('Time [seconds]')
		ax2.axis('auto')

		basename = filename[filename.rfind('/')+1:]
		output = basename.replace('.txt', '.png')
		output = output.replace('.dat', '.png')
		fig.savefig(output)
	

if __name__ == "__main__":
	main(sys.argv[1:])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Randomly select 20 antennae from LWA-1 and plot the uv-plane coverage for 
a zenith snapshot and the expected beam."""

import os
import sys
import math
import numpy

import lsl.correlator.uvUtils as uvUtils
import lsl.common.stations as lwa_common

import matplotlib.pyplot as plt
from matplotlib.ticker import NullFormatter

allStands = numpy.arange(1,257)
# List of installed stands from Joe
installedStands = numpy.array([13, 78, 9, 38, 34, 70, 106, 10, 170, 142, 32, 8, 71, 33, 69, 168, 108, 
						140, 110, 76, 161, 101, 99, 31, 67, 159, 185, 97, 127, 157, 125, 208, 
						123, 181, 206, 183, 153, 174, 225, 176, 172, 204, 178, 200, 202, 228, 
						210, 212, 214, 187, 15, 80, 14, 148, 112, 46, 118])
# List of stands we *do not* want to use.  204, 202, 178, 148, 176, 225, and 
# 200 are close to the shelter and may have some self-RFI problems.
excludeStands = numpy.array([204, 202, 178, 148, 176, 225, 200])

def shift(data, xy=[0,0]):
	data1 = numpy.concatenate([data[xy[0]:], data[:xy[0]]], axis=0)
	data2 = numpy.concatenate([data1[:,xy[1]:], data1[:,:xy[1]]], axis=1)

	return data2


def randomSelection(N, readyStands, allStands, Fraction=0.75, IncludeOutlier=False):
	Nready = int(Fraction*N)
	Nall = N - Nready
	
	standList = []
	for i in range(Nready):
		randInt = numpy.random.randint(0, readyStands.shape[0])
		while readyStands[randInt] in standList or readyStands[randInt] in excludeStands:
			randInt = numpy.random.randint(0, readyStands.shape[0])
		standList.append( readyStands[randInt] )
	
	for i in range(Nall):
		randInt = numpy.random.randint(0, allStands.shape[0])
		while allStands[randInt] in standList or allStands[randInt] in excludeStands or allStands[randInt] in readyStands:
			randInt = numpy.random.randint(0, allStands.shape[0])
		standList.append( allStands[randInt] )

	if IncludeOutlier:
		standList[0] = -1

	return numpy.array(standList)


def main(args):
	# Set the LWA Station
	station = lwa_common.lwa1()

	if len(args) == 0:
		fraction = 0.75
		stands = randomSelection(20, installedStands, allStands, Fraction=fraction, IncludeOutlier=True)
	elif len(args) > 1:
		stands = numpy.array([int(i) for i in args])
	else:
		fraction = float(args[0])
		stands = randomSelection(20, installedStands, allStands, Fraction=fraction, IncludeOutlier=True)

	print stands.shape[0], stands
	print "New Stands:"
	for stand in stands:
		if stand not in installedStands:
			print " %3i" % stand

	HA = 0.0
	dec = station.lat*180.0/math.pi
	freq = 50.0e6

	uvw = uvUtils.computeUVW(stands, HA=HA, dec=dec, freq=freq)
	uvw = numpy.squeeze(uvw[:,:,0])
	uvw = uvw * 2.9979245800e8 / freq

	# Coursely grid the uv data to come up with a rough beam
	if min(stands) < 0:
		grid = numpy.zeros((900,900))
		for i in range(uvw.shape[0]):
			u = round(uvw[i,0])+450
			v = round(uvw[i,1])+450
			grid[u,v] = grid[u,v]+1
	else:
		grid = numpy.zeros((240,240))
		for i in range(uvw.shape[0]):
			u = round(uvw[i,0])+120
			v = round(uvw[i,1])+120
			grid[u,v] = grid[u,v]+1

	# Plot
	fig = plt.figure(figsize=(8,8))
	ax1 = plt.axes([0.30, 0.30, 0.60, 0.60])
	ax2 = plt.axes([0.30, 0.05, 0.60, 0.15])
	ax3 = plt.axes([0.05, 0.30, 0.15, 0.60])
	#ax4 = plt.axes([0.05, 0.05, 0.10, 0.15])
	ax4 = plt.axes([0.08, 0.08, 0.15, 0.15])
	ax5 = plt.axes([0.32, 0.32, 0.15, 0.15])

	beam = numpy.fft.fft2(grid)
	if min(stands) < 0:
		beam = shift(beam, xy=(450,450))
		ax5.imshow((numpy.log10(beam[375:525, 375:525])*10.0).real, interpolation="nearest")
		ax5.xaxis.set_major_formatter( NullFormatter() )
		ax5.yaxis.set_major_formatter( NullFormatter() )
	else:
		beam = shift(beam, xy=(120,120))
		ax5.imshow((numpy.log10(beam.real[100:140, 100:140])*10.0).real, interpolation="nearest")
		ax5.xaxis.set_major_formatter( NullFormatter() )
		ax5.yaxis.set_major_formatter( NullFormatter() )

	c = ax1.scatter(uvw[:,0], uvw[:,1], c=uvw[:,2], s=40.0, alpha=0.75)
	d = ax1.scatter(-uvw[:,0], -uvw[:,1], c=-uvw[:,2], s=40.0, alpha=0.75)
	ax1.set_xlabel('u [m]')
	ax1.set_ylabel('v [m]')
	ax1.set_title('UV Coverage for HA=%+.3f$^h$, $\delta$=%+.3f$^\circ$ at %s' % (HA, dec, station.name))

	ax2.scatter(uvw[:,0], uvw[:,2], c=uvw[:,2], s=40.0)
	ax2.scatter(-uvw[:,0], -uvw[:,2], c=-uvw[:,2], s=40.0)
	ax2.xaxis.set_major_formatter( NullFormatter() )
	ax2.set_ylabel('w')

	ax3.scatter(uvw[:,2], uvw[:,1], c=uvw[:,2], s=40.0)
	ax3.scatter(-uvw[:,2], -uvw[:,1], c=-uvw[:,2], s=40.0)
	ax3.yaxis.set_major_formatter( NullFormatter() )
	ax3.set_xlabel('w')

	#cb = plt.colorbar(c, cax=ax4, orientation='vertical')

	rad = numpy.zeros(uvw.shape[0])
	for i in range(rad.shape[0]):
		rad[i] = math.sqrt( uvw[i,0]**2.0 + uvw[i,1]**2.0 + uvw[i,2]**2.0 ) * freq / 2.9979245800e8
	ax4.hist(rad, 20)
	ax4.set_xlabel('uv Radius [$\lambda$]')
	ax4.set_ylabel('Baselines')
	if min(stands) < 0:
		ax4.set_xticks([0, 15, 30, 45, 60, 75])
		ax4.set_yticks([0, 10, 20, 30, 40, 50])
	else:
		ax4.set_xticks([0, 4, 8, 12, 16, 20])
		ax4.set_yticks([0, 4, 8, 12, 16, 20])

	if min(stands) < 0:
		ax1.set_xlim([-450, 450])
		ax1.set_ylim([-450, 450])
	else:
		ax1.set_xlim([-120, 120])
		ax1.set_ylim([-120, 120])
	ax2.set_xlim( ax1.get_xlim() )
	ax3.set_ylim( ax1.get_ylim() )

	if min(stands) < 0:
		ax1.text(465, 375, 'Stands:')
		for i in range(stands.shape[0]):
			if stands[i] > 0:
				if stands[i] in installedStands:
					ax1.text(476.25, 337.5-i*37.5, "%i" % stands[i])
				else:
					ax1.text(476.25, 337.5-i*37.5, "%i (*)" % stands[i])
			else:
				ax1.text(476.25, 337.5-i*37.5, 'outlier')
	else:
		ax1.text(124, 100, 'Stands:')
		for i in range(stands.shape[0]):
			ax1.text(127, 90-i*10, str(stands[i]))

	plt.show()
	fig.savefig('uv-plane.png')


if __name__ == "__main__":
	main(sys.argv[1:])



# -*- coding: utf-8 -*-

"""Modules to take DRX/TBW/TBN data and write it to a SDFITS file.

.. warning:
	The FITS files created by this module to not strictly conform to the
	SDFITS convention.
"""

import os
import sys
import numpy
import pyfits
from datetime import datetime, timedelta, tzinfo

from lsl.common import dp as dp_common
from lsl.common.warns import warnDeprecated
from lsl.correlator import fx as correlate

__version__ = '0.3'
__revision__ = '$ Revision: 15 $'
__all__ = ['SDFITS', 'TBW', 'TBN', 'DRX', '__version__', '__revision__', '__all__']


class SDFITS(object):
	"""Class that holds TSFITS data until it is ready to be writen to disk."""

	def __init__(self, filename, mode, LFFT=128, Overwrite=False, UseQueue=True, verbose=False):
		"""Initialize a SDFITS object using a filename, an observation mode (TBW,  
		TBN, or DRX), and a FFT length in channels.  Optionally, SDFITS can be 
		told to overwrite the file if it already exists using the 'Overwrite' keyword."""

		assert(mode in ['TBW', 'TBN', 'DRX'])

		self.filename = filename
		self.mode = mode
		self.hdulist = None
		self.fftLength = LFFT
		self.UseQueue = UseQueue
		self.queue = {}
		self.queueLimit = 10000
		self.site = 'LWA'
		self.sampleRate = dp_common.fS
		self.firstSamples = {}
		self.verbose = verbose

		if os.path.exists(filename) and not Overwrite:
			self.hdulist = pyfits.open(self.filename, mode="update", memmap=0)
			self.standCount = self.hdulist[0].header['nstand']
		else:
			if os.path.exists(filename):
				os.path.delete(filename)

			self.standCount = 0
			primary = pyfits.PrimaryHDU()
			primary.header.update('TELESCOP', self.site)
			primary.header.update('NSTAND', self.standCount)

			hdulist = pyfits.HDUList([primary])
			hdulist.writeto(filename)

			self.hdulist = pyfits.open(self.filename, mode="update", memmap=0)

	def info(self):
		"""Short-cut to the pyfits.info() function on an opened FITS file."""

		self.hdulist.info()

	def flush(self):
		"""Short-cut to the pyfits.flush() function on an opened FITS file."""

		self.hdulist.flush()

	def close(self):
		"""Empty the data queue (if it exists) and write all changes to disk 
		using flush."""

		if self.UseQueue:
			self.__emptyQueue()
		self.flush()

	def setSite(self, site):
		"""Set the TELESCOP keyword in the primary HDU using an lsl.common.stations
		object."""

		self.site = site.name

		for hdu in self.hdulist:
			hdu.header.update('TELESCOP', self.site)

		self.flush()

	def setSampleRate(self, sampleRate):
		"""Set the sample rate in Hz of the data for TBN and DRX observations."""

		if self.mode != 'TBW':
			self.sampleRate = sampleRate

	def setCentralFrequency(self, centralFreq):
		"""Set the central frequency in Hz of the data for TBN and DRX 
		observations."""

		if self.mode != 'TBW':
			self.centralFreq = centralFreq

	def __findExtension(self, stand):
		"""Private function to find out which extension stores the stand in 
		question.  None is returned is that stand is not in the SDFITS file."""

		extension = None

		i = 1
		for hdu in self.hdulist[1:]:
			if int(hdu.header['stand']) == int(stand):
				extension = i
				break
			i = i + 1

		return extension

	def __makeAppendTable(self, extension, AddRows=1):
		"""Private function to make a temporary table for appending data."""

		nrows = self.hdulist[extension].data.shape[0]
		tempHDU = pyfits.new_table(self.hdulist[extension].columns, nrows=nrows+AddRows)
		tempHDU.header = self.hdulist[extension].header.copy()
	
		return tempHDU

	def __applyAppendTable(self, extension, tempHDU):
		"""Private function to replace the given extension with the temporary
		table."""

		self.hdulist[extension] = tempHDU
		self.flush()

	def __addDataSingle(self, frame):
		"""Private function to add a single data entry to a SDFITS file.  This 
		method is not particular fast since the existing file needs to be copied
		to append data to the end of a particular extension."""

		if self.mode == 'TBW':
			stand = frame.parseID()
		else:
			stand, pol = frame.parseID()
		extension = self.__findExtension(stand)

		if extension is None:
			if self.verbose:
				print "Stand '%i' not found, creating new binary table extension" % stand
			self.standCount = self.standCount + 1

			if self.mode == 'TBW':
				freq, framePS = correlate.calcSpectra(frame.data.xy, LFFT=self.fftLength, SampleRate=self.sampleRate)

				# Data - power spectrum
				c1 = pyfits.Column(name='data', format='%iD' % (self.fftLength-1), array=framePS.astype(numpy.float_))
				# Polarization
				c2 = pyfits.Column(name='pol', format='1I', array=numpy.array([0, 1], dtype=numpy.int16))
				# Time
				c3 = pyfits.Column(name='time', format='1D', array=numpy.array([0, 0], dtype=numpy.int16))
			else:
				data = frame.data.iq
				data.shape = (1,data.shape[0])
				freq, framePS = correlate.calcSpectra(data, LFFT=self.fftLength, SampleRate=self.sampleRate)
				framePS.shape = (1,framePS.shape[0])

				# Data - power spectrum
				c1 = pyfits.Column(name='data', format='%iD' % (self.fftLength-1), array=framePS.astype(numpy.float_))
				# Polarization
				c2 = pyfits.Column(name='pol', format='1I')
				# Time
				c3 = pyfits.Column(name='time', format='1D')

			# Define the collection of columns
			colDefs = pyfits.ColDefs([c1, c2, c3])

			# Get the time of the first sample and convert it to a datetime
			self.firstSamples[stand] = long(frame.data.timeTag / dp_common.fS)
			firstSample = datetime.utcfromtimestamp(self.firstSamples[stand])

			sdfits = pyfits.new_table(colDefs)
			sdfits.header.update('EXTNAME', 'SINGLE DISH', after='tfields')
			sdfits.header.update('EXTVER', self.standCount, after='EXTNAME')
			sdfits.header.update('STAND', stand, after='EXTVER')

			# Define static fields - core
			sdfits.header.update('OBJECT', 'zenith')
			sdfits.header.update('TELESCOP', self.site)
			sdfits.header.update('BANDWID', 78.0e6)
			sdfits.header.update('DATE-OBS', firstSample.isoformat('T'))
			sdfits.header.update('EXPOSURE', 0.0)
			sdfits.header.update('TSYS', 1.0)
			# Define static fields - virtual columns
			sdfits.header.update('CTYPE1', 'FREQ-OBS')
			sdfits.header.update('CRVAL1', freq[0])
			sdfits.header.update('CRPIX1', 1)
			sdfits.header.update('CDELT1', (freq[1]-freq[0]))
			sdfits.header.update('CTYPE2', 'HA')
			sdfits.header.update('CRVAL2', 0.0)
			sdfits.header.update('CUNIT2', 'degrees')
			sdfits.header.update('CTYPE3', 'DEC')
			sdfits.header.update('CRVAL3', 0.0)
			sdfits.header.update('CUNIT3', 'degrees')
			sdfits.header.update('OBSMODE', self.mode)
			sdfits.header.update('NCHAN', len(freq))

			self.hdulist.append(sdfits)
			self.hdulist[0].header.update('NSTAND', self.standCount)
			self.flush()
			
			self.hdulist[-1].data.field('pol')[0] = 0
			self.hdulist[-1].data.field('time')[0] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
			if self.mode == 'TBW':
				self.hdulist[-1].data.field('pol')[1] = 1
				self.hdulist[-1].data.field('time')[1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
		else:
			# Make sure that we have a first sample time to reference to if 
			# we are adding on to the end of a file
			if stand not in self.firstSamples.keys():
				firstSample = datetime.strptime(self.hdulist[extension].header['DATE-OBS'], "%Y-%m-%dT%H:%M:%S")
				self.firstSamples[stand] = long(time.mktime(firstSample.timetuple()))

			nrows = self.hdulist[extension].data.shape[0]
			if self.mode == 'TBW':
				freq, framePS = correlate.calcSpectra(frame.data.xy, LFFT=self.fftLength, SampleRate=self.sampleRate)

				tempHDU = self.__makeAppendTable(extension, AddRows=2)
				tempHDU.data.field('data')[nrows] = numpy.squeeze(framePS.astype(numpy.float_)[0,:])
				tempHDU.data.field('pol')[nrows] = numpy.array([0], dtype=numpy.int16)
				tempHDU.data.field('time')[nrows] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
				tempHDU.data.field('data')[nrows+1] = numpy.squeeze(framePS.astype(numpy.float_)[1,:])
				tempHDU.data.field('pol')[nrows+1] = numpy.array([1], dtype=numpy.int16)
				tempHDU.data.field('time')[nrows+1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
			else:
				data = frame.data.iq
				data.shape = (1,data.shape[0])
				freq, framePS = correlate.calcSpectra(data, LFFT=self.fftLength, SampleRate=self.sampleRate)
				framePS.shape = (1,framePS.shape[0])

				tempHDU = self.__makeAppendTable(extension, AddRows=1)
				tempHDU.data.field('data')[nrows:] = framePS.astype(numpy.float_)
				tempHDU.data.field('pol')[nrows:] = numpy.array([pol], dtype=numpy.int16)
				tempHDU.data.field('time')[nrows:] = numpy.array([frame.data.timeTag]) / dp_common.fS - self.firstSamples[stand]

			self.__applyAppendTable(extension, tempHDU)

		self.flush()

	def __addDataQueue(self, frame):
		"""Private function similar to __addDataSignle, but it saves the data to 
		memory (self.queue) to be written once the queue fills up."""

		if self.mode == 'TBW':
			stand = frame.parseID()
		else:
			stand, pol = frame.parseID()
		if stand not in self.queue.keys():
			self.queue[stand] = []
		self.queue[stand].append(frame)

		if len(self.queue[stand]) >= self.queueLimit:
			start = 0
			extension = self.__findExtension(stand)

			if extension is None:
				if self.verbose:
					print "Stand '%i' not found, creating new binary table extension" % stand
				self.standCount = self.standCount + 1

				if self.mode == 'TBW':
					freq, framePS = correlate.calcSpectra(frame.data.xy, LFFT=self.fftLength, SampleRate=self.sampleRate)

					# Data - power spectrum
					c1 = pyfits.Column(name='data', format='%iD' % (self.fftLength-1), array=framePS.astype(numpy.float_))
					# Polarization
					c2 = pyfits.Column(name='pol', format='1I', array=numpy.array([0, 1], dtype=numpy.int16))
					# Time
					c3 = pyfits.Column(name='time', format='1D', array=numpy.array([0, 0], dtype=numpy.int16))
				else:
					data = frame.data.iq
					data.shape = (1,data.shape[0])
					freq, framePS = correlate.calcSpectra(data, LFFT=self.fftLength, SampleRate=self.sampleRate)
					framePS.shape = (1,framePS.shape[0])

					# Data - power spectrum
					c1 = pyfits.Column(name='data', format='%iD' % (self.fftLength-1), array=framePS.astype(numpy.float_))
					# Polarization
					c2 = pyfits.Column(name='pol', format='1I')
					# Time
					c3 = pyfits.Column(name='time', format='1D')

				# Define the collection of columns
				colDefs = pyfits.ColDefs([c1, c2, c3])

				# Get the time of the first sample and convert it to a datetime
				self.firstSamples[stand] = long(frame.data.timeTag / dp_common.fS)
				firstSample = datetime.utcfromtimestamp(self.firstSamples[stand])

				sdfits = pyfits.new_table(colDefs)
				sdfits.header.update('EXTNAME', 'SINGLE DISH', after='tfields')
				sdfits.header.update('EXTVER', self.standCount, after='EXTNAME')
				sdfits.header.update('STAND', stand, after='EXTVER')

				# Define static fields - core
				sdfits.header.update('OBJECT', 'zenith')
				sdfits.header.update('TELESCOP', self.site)
				sdfits.header.update('BANDWID', 78.0e6)
				sdfits.header.update('DATE-OBS', firstSample.isoformat('T'))
				sdfits.header.update('EXPOSURE', 0.0)
				sdfits.header.update('TSYS', 1.0)
				# Define static fields - virtual columns
				sdfits.header.update('CTYPE1', 'FREQ-OBS')
				sdfits.header.update('CRVAL1', freq[0])
				sdfits.header.update('CRPIX1', 1)
				sdfits.header.update('CDELT1', (freq[1]-freq[0]))
				sdfits.header.update('CTYPE2', 'HA')
				sdfits.header.update('CRVAL2', 0.0)
				sdfits.header.update('CUNIT2', 'degrees')
				sdfits.header.update('CTYPE3', 'DEC')
				sdfits.header.update('CRVAL3', 0.0)
				sdfits.header.update('CUNIT3', 'degrees')
				sdfits.header.update('OBSMODE', self.mode)
				sdfits.header.update('NCHAN', len(freq))

				self.hdulist.append(sdfits)
				self.hdulist[0].header.update('NSTAND', self.standCount)
				self.flush()
				
				self.hdulist[-1].data.field('pol')[0] = 0
				self.hdulist[-1].data.field('time')[0] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
				if self.mode == 'TBW':
					self.hdulist[-1].data.field('pol')[1] = 1
					self.hdulist[-1].data.field('time')[1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]

				self.flush()
				extension = self.__findExtension(stand)

			# Make sure that we have a first sample time to reference to if 
			# we are adding on to the end of a file
			if stand not in self.firstSamples.keys():
				firstSample = long(datetime.strptime(self.hdulist[extension].header['DATE-OBS'], "%Y-%m-%dT%H:%M:%S"))
				self.firstSamples[stand] = time.mktime(firstSample.timetuple())

			nrows = self.hdulist[extension].data.shape[0]
			if self.mode == 'TBW':
				tempHDU = self.__makeAppendTable(extension, AddRows=2*(self.queueLimit-start))
				for count,frame in zip(range(len(self.queue[stand][start:])), self.queue[stand][start:]):
					freq, framePS = correlate.calcSpectra(frame.data.xy, LFFT=self.fftLength)

					tempHDU.data.field('data')[nrows+2*count] = numpy.squeeze(framePS.astype(numpy.float_)[0,:])
					tempHDU.data.field('pol')[nrows+2*count] = 0
					tempHDU.data.field('time')[nrows+2*count] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
					tempHDU.data.field('data')[nrows+2*count+1] = numpy.squeeze(framePS.astype(numpy.float_)[1,:])
					tempHDU.data.field('pol')[nrows+2*count+1] = 1
					tempHDU.data.field('time')[nrows+2*count+1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
			else:
				tempHDU = self.__makeAppendTable(extension, AddRows=(self.queueLimit-start))
				for count,frame in zip(range(len(self.queue[stand][start:])), self.queue[stand][start:]):
					data = frame.data.iq
					data.shape = (1,data.shape[0])
					freq, framePS = correlate.calcSpectra(data, LFFT=self.fftLength, SampleRate=self.sampleRate)
					framePS.shape = (1,framePS.shape[0])

					stand,pol = frame.parseID()
					tempHDU.data.field('data')[nrows+2*count] = numpy.squeeze(framePS.astype(numpy.float_)[0,:])
					tempHDU.data.field('pol')[nrows+2*count] = pol
					tempHDU.data.field('time')[nrows+2*count] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]

			self.__applyAppendTable(extension, tempHDU)
			del(self.queue[stand])

	def __emptyQueue(self):
		"""Private function to empty a empty the self.queue dictionary on demand
		if it exists."""

		for stand in self.queue.keys():
			if len(self.queue[stand]) == 0:
				continue
			queueSize = len(self.queue[stand])
			start = 0
			extension = self.__findExtension(stand)

			if extension is None:
				start = 1
				frame = self.queue[stand][0]
				
				if self.verbose:
					print "Stand '%i' not found, creating new binary table extension" % stand
				self.standCount = self.standCount + 1

				if self.mode == 'TBW':
					freq, framePS = correlate.calcSpectra(frame.data.xy, LFFT=self.fftLength, SampleRate=self.sampleRate)

					# Data - power spectrum
					c1 = pyfits.Column(name='data', format='%iD' % (self.fftLength-1), array=framePS.astype(numpy.float_))
					# Polarization
					c2 = pyfits.Column(name='pol', format='1I', array=numpy.array([0, 1], dtype=numpy.int16))
					# Time
					c3 = pyfits.Column(name='time', format='1D', array=numpy.array([0, 0], dtype=numpy.int16))
				else:
					data = frame.data.iq
					data.shape = (1,data.shape[0])
					freq, framePS = correlate.calcSpectra(data, LFFT=self.fftLength, SampleRate=self.sampleRate)

					# Data - power spectrum
					c1 = pyfits.Column(name='data', format='%iD' % (self.fftLength-1), array=framePS.astype(numpy.float_))
					# Polarization
					c2 = pyfits.Column(name='pol', format='1I')
					# Time
					c3 = pyfits.Column(name='time', format='1D')

				# Define the collection of columns
				colDefs = pyfits.ColDefs([c1, c2, c3])

				# Get the time of the first sample and convert it to a datetime
				self.firstSamples[stand] = long(frame.data.timeTag / dp_common.fS)
				firstSample = datetime.utcfromtimestamp(self.firstSamples[stand])

				# Get the time of the first sample and convert it to a datetime
				self.firstSamples[stand] = long(frame.data.timeTag / dp_common.fS)
				firstSample = datetime.utcfromtimestamp(self.firstSamples[stand])

				sdfits = pyfits.new_table(colDefs)
				sdfits.header.update('EXTNAME', 'SINGLE DISH', after='tfields')
				sdfits.header.update('EXTVER', self.standCount, after='EXTNAME')
				sdfits.header.update('STAND', stand, after='EXTVER')

				# Define static fields - core
				sdfits.header.update('OBJECT', 'zenith')
				sdfits.header.update('TELESCOP', 'LWA-1')
				sdfits.header.update('BANDWID', 78.0e6)
				sdfits.header.update('DATE-OBS', firstSample.isoformat('T'))
				sdfits.header.update('EXPOSURE', 0.0)
				sdfits.header.update('TSYS', 1.0)
				# Define static fields - virtual columns
				sdfits.header.update('CTYPE1', 'FREQ-OBS')
				sdfits.header.update('CRVAL1', freq[0])
				sdfits.header.update('CRPIX1', 1)
				sdfits.header.update('CDELT1', (freq[1]-freq[0]))
				sdfits.header.update('CTYPE2', 'HA')
				sdfits.header.update('CRVAL2', 0.0)
				sdfits.header.update('CUNIT2', 'degrees')
				sdfits.header.update('CTYPE3', 'DEC')
				sdfits.header.update('CRVAL3', 0.0)
				sdfits.header.update('CUNIT3', 'degrees')
				sdfits.header.update('OBSMODE', self.mode)
				sdfits.header.update('NCHAN', len(freq))

				self.hdulist.append(sdfits)
				self.hdulist[0].header.update('NSTAND', self.standCount)
				self.flush()
				
				self.hdulist[-1].data.field('pol')[0] = 0
				self.hdulist[-1].data.field('time')[0] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
				if self.mode == 'TBW':
					self.hdulist[-1].data.field('pol')[1] = 1
					self.hdulist[-1].data.field('time')[1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]

				self.flush()
				extension = self.__findExtension(stand)

			# Make sure that we have a first sample time to reference to if 
			# we are adding on to the end of a file
			if stand not in self.firstSamples.keys():
				firstSample = long(datetime.strptime(self.hdulist[extension].header['DATE-OBS'], "%Y-%m-%dT%H:%M:%S"))
				self.firstSamples[stand] = time.mktime(firstSample.timetuple())

			nrows = self.hdulist[extension].data.shape[0]
			if self.mode == 'TBW':
				tempHDU = self.__makeAppendTable(extension, AddRows=2*(queueSize-start))
				for count,frame in zip(range(len(self.queue[stand][start:])), self.queue[stand][start:]):
					freq, framePS = correlate.calcSpectra(frame.data.xy, LFFT=self.fftLength)

					tempHDU.data.field('data')[nrows+2*count] = numpy.squeeze(framePS.astype(numpy.float_)[0,:])
					tempHDU.data.field('pol')[nrows+2*count] = 0
					tempHDU.data.field('time')[nrows+2*count] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
					tempHDU.data.field('data')[nrows+2*count+1] = numpy.squeeze(framePS.astype(numpy.float_)[1,:])
					tempHDU.data.field('pol')[nrows+2*count+1] = 1
					tempHDU.data.field('time')[nrows+2*count+1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
			else:
				tempHDU = self.__makeAppendTable(extension, AddRows=(queueSize-start))
				for count,frame in zip(range(len(self.queue[stand][start:])), self.queue[stand][start:]):
					data = frame.data.iq
					data.shape = (1,data.shape[0])
					freq, framePS = correlate.calcSpectra(data, LFFT=self.fftLength, SampleRate=self.sampleRate)
					framePS.shape = (1,framePS.shape[0])

					stand,pol = frame.parseID()
					tempHDU.data.field('data')[nrows+2*count] = numpy.squeeze(framePS.astype(numpy.float_)[0,:])
					tempHDU.data.field('pol')[nrows+2*count] = pol
					tempHDU.data.field('time')[nrows+2*count] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]

			self.__applyAppendTable(extension, tempHDU)
			del(self.queue[stand])

	def addStandData(self, frame):
		"""Add a frame object to the SDFITS file.  This function takes care of 
		figuring out which extension the data goes to and how it should be formated.
		This function also updates the primary HDU with information about the data, 
		i.e., TBW data bits."""

		if self.mode == 'TBW':
			try:
				self.hdulist[0].header['TBWBITS']
			except:
				self.hdulist[0].header.update('TBWBITS', frame.getDataBits())
				self.flush()

		if self.UseQueue:
			self.__addDataQueue(frame)
		else:
			self.__addDataSingle(frame)

	def getStandData(self, stand):
		"""Retrieve a dictionary of all data stored for a particular stand.  The 
		dictionary keys are:
		  * *data* - numpy array of data
		  * *pol* - numpy array of polarizations
		  * *time* - numpy array of times in samples at f_S since DATE-OBS
		"""

		extension = self.__findExtension(stand)
		
		if extension is None:
			output = {}
			for key in ['data', 'pol', 'time']:
				output[key] = None
		else:
			output = {}
			for key in ['data', 'pol', 'time']:
				output[key] = self.hdulist[extension].data.field(key)
		return output


class TBW(SDFITS):
	"""Sub-class of SDFITS for dealing with TBW data in particular."""

	def __init__(self, filename, LFFT=128, Overwrite=False, UseQueue=True, verbose=False):
		super(TBW, self).__init__(filename, 'TBW', LFFT=LFFT, Overwrite=Overwrite, UseQueue=UseQueue, verbose=verbose)


class TBN(SDFITS):
	"""Sub-class of SDFITS for dealing with TBN data in particular."""

	def __init__(self, filename, LFFT=128, Overwrite=False, UseQueue=True, verbose=False):
		super(TBN, self).__init__(filename, 'TBN', LFFT=LFFT, Overwrite=Overwrite, UseQueue=UseQueue, verbose=verbose)


#class DRX(SDFITS):
	#"""Sub-class of SDFITS for dealing with DRX data in particular."""

	#def __init__(self, filename, Overwrite=False, UseQueue=True, verbose=False):
		#super(DRX, self).__init__(filename, 'DRX', LFFT=4096, Overwrite=Overwrite, UseQueue=UseQueue, verbose=verbose)



# -*- coding: utf-8 -*-

"""Module to allow for post-acquisition delay-and-sum beamforming with TBW 
and TBN data using both integer sample delays for time series data 
(intDelayAndSum) and fractional delays for time series data transformed to 
the frequency domain (fftDelayAndSum).  fftDelayAndSum is still under 
development."""

import os
import sys
import aipy
import math
import numpy

from lsl.common.paths import data as dataPath
from lsl.common.constants import c
from lsl.common import dp as dp_common
from lsl.correlator import uvUtils

__version__ = '0.2'
__revision__ = '$ Revision: 7 $'
__all__ = ['BeamformingError', 'calcDelay', 'intDelayAndSum', 'intBeamShape', 'fftDelayAndSum', 'fftBeamShape', '__version__', '__revision__', '__all__']

class BeamformingError(Exception):
	"""Base class for all exceptions in this file."""

	def __init__(self, strerror, errno='-1'):
		self.errno = errno
		self.strerror = strerror
		self.filename = None
		self.args = (errno, strerror)

	def __str__(self):
		return "%s" % self.strerror


def __loadStandResponse(freq=49.0e6):
	"""Create an aipy.amp.beam object that holds the reponse for a single 
	isolated stand.  The stand response is based on NEC4 models at a variety
	of frequencies within the LWA frequency range."""

	# Read in the spherical harmonic representation of the beam distributed with
	# LSL
	dd = numpy.load(os.path.join(dataPath, 'beam-shape.npz'))
	coeffs = dd['coeffs']

	# Calculate how many harmonics are stored in the data set and reorder the data
	# to aipy's liking
	deg = coeffs.shape[0]-1
	lmax = int((math.sqrt(1+8*coeffs.shape[1])-3)/2)
	beamShapeDict = {}
	for i in range(deg+1):
		beamShapeDict[i] = numpy.squeeze(coeffs[-1-i,:])

	# Build the beam object and done
	return aipy.amp.BeamAlm(numpy.array([freq/1e9]), lmax=lmax, mmax=lmax, deg=deg, nside=128, coeffs=beamShapeDict)


def calcDelay(stands, freq=49.0e6, azimuth=0.0, elevation=90.0):
	"""Calculate the time delays for delay-and-sum beam forming a collection of 
	stands looking in at a particular azimuth and elevation (both in degrees).  
	A numpy array of the needed delays in seconds is returned."""

	# Make sure the pointing coordinates make sense
	if elevation < 0 or elevation > 90:
		raise BeamformingError("Pointing elevation (%.2f deg) is out of range [0, 90]" % elevation)
	if azimuth < 0 or azimuth > 360:
		raise BeamformingError("Pointing azimuth (%.2f deg) is out of range [0, 360]" % azimuth)

	# Get the positions of the stands and compute the mean center of the array
	xyz = uvUtils.getXYZ(stands)
	arrayX = xyz[:,0].mean()
	arrayY = xyz[:,1].mean()
	arrayZ = xyz[:,2].mean()

	# Build up a unit vector that points in the direction azimuth,elevation
	rAz = azimuth*numpy.pi/180.0
	rEl = elevation*numpy.pi/180.0
	source = numpy.array([numpy.cos(rEl)*numpy.sin(rAz), 
						numpy.cos(rEl)*numpy.cos(rAz), 
						numpy.sin(rEl)])

	# Compute the stand positions relative to the average and loop over stands
	# to compute the time delays in seconds
	arrayXYZ = xyz - numpy.array([arrayX, arrayY, arrayZ])
	delays = numpy.zeros((len(stands),))
	for i in list(range(len(stands))):
		delays[i] = numpy.dot(source, arrayXYZ[i,:]) / c

	# Get the cable delays for each stand and add that in as well
	dlyCache = uvUtils.CableCache(freq, applyDispersion=True)
	for i in list(range(len(stands))):
		delays[i] = dlyCache.cableDelay(stands[i]) - delays[i]

	# Done
	return delays


def intDelayAndSum(stands, data, sampleRate=dp_common.fS, azimuth=0.0, elevation=90.0):
	"""Given a list of stands and a data stream of the form stands x times, 
	delay and sum the data stream into one beam.  The delays applied are 
	integer sample delays.  Return a numpy array of the time series data 
	associated with the formed beam."""

	# Get the stand delays and convert the delay times from seconds to samples
	delays = calcDelay(stands, azimuth=azimuth, elevation=elevation)
	delays = numpy.round(delays*sampleRate).astype(numpy.int16)

	# Make the delays into something meaningful for the shifting of the data 
	# streams
	delays = delays.max() - delays

	# Delay and sum by looping over stands inside of looping over times
	output = numpy.zeros((data.shape[1]-delays.max()), dtype=data.dtype)
	for s in list(range(len(stands))):
		start = delays[s]
		stop = data.shape[1] - delays.max() + start
		output = output + data[s,start:stop]

	# Done
	return output


def __intBeepAndSweep(stands, arrayXYZ, t, azimuth, elevation, dlyCache=None, beamShape=1.0, sampleRate=dp_common.fS, direction=(0.0, 90.0)):
	"""Worker function for intBeamShape that 'beep's (makes a simulated signals) and
	'sweep's (delays it appropriately)."""

	# Convert from degrees to radian
	rAz = azimuth*numpy.pi/180.0
	rEl = elevation*numpy.pi/180.0

	# Unit vector for the currect on-sky location
	currPos = numpy.array([numpy.cos(rEl)*numpy.sin(rAz), 
						numpy.cos(rEl)*numpy.cos(rAz), 
						numpy.sin(rEl)])
	# Stand response in this direction
	currResponse = beamShape

	# Loop over stands to build the simulated singnals
	signals = numpy.zeros((len(stands), len(t)))
	for i in list(range(len(stands))):
		currDelay = dlyCache.cableDelay(stands[i]) - numpy.dot(currPos, arrayXYZ[i,:]) / c
		signals[i,:] = currResponse * numpy.cos(2*numpy.pi*49.0e6*(t + currDelay))

	# Beamform with delay-and-sum and store the RMS result
	beamHere = intDelayAndSum(stands, signals, sampleRate=sampleRate, azimuth=direction[0], elevation=direction[1])

	# Return
	sigHere = numpy.sqrt((beamHere**2).mean())
	return sigHere


def intBeamShape(stands, sampleRate=dp_common.fS, azimuth=0.0, elevation=90.0, progress=False, DisablePool=False):
	"""Given a list of stands, compute the on-sky response of the delay-and-sum
	scheme implemented in intDelayAndSum.  A 360x90 numpy array spaning azimuth
	and elevation is returned."""

	# Get the stand delays and convert the delay times from seconds to samples
	delays = calcDelay(stands, freq=49.0e6, azimuth=azimuth, elevation=elevation)
	delays = numpy.round(delays*sampleRate).astype(numpy.int16)

	# Build up a base time array, load in the cable delays, and get the stand 
	# positions for geometric delay calculations.
	t = numpy.arange(0,1000)/sampleRate
	dlyCache = uvUtils.CableCache(49.0e6, applyDispersion=True)
	xyz = uvUtils.getXYZ(stands)
	arrayX = xyz[:,0].mean()
	arrayY = xyz[:,1].mean()
	arrayZ = xyz[:,2].mean()
	arrayXYZ = xyz - numpy.array([arrayX, arrayY, arrayZ])

	# Load in the respoonse of a single isolated stand
	standBeam = __loadStandResponse(freq=49.0e6)

	# The processing module allows for the creation of worker pools to help speed
	# things along.  If the processing module is found, use it.  Otherwise, set
	# the 'usePool' variable to false and run single threaded.
	try:
		from processing import Pool
		
		# To get results pack from the pool, you need to keep up with the workers.  
		# In addition, we need to keep up with which workers goes with which 
		# baseline since the workers are called asychronisly.  Thus, we need a 
		# taskList array to hold tuples of baseline ('count') and workers.
		taskPool = Pool(processes=4)
		taskList = []

		usePool = True
		progress = False
	except ImportError:
		usePool = False

	# Turn off the thread pool if we are explicitly told not to use it.
	if DisablePool:
		usePool = False

	# Build up the beam shape over all azimuths and elevations
	beamShape =  numpy.zeros((360,90))
	for az in list(range(360)):
		rAz = az*numpy.pi/180.0
		for el in list(range(90)):
			rEl = el*numpy.pi/180.0
			beamShape[az,el] = standBeam.response(aipy.coord.azalt2top(numpy.concatenate([[rAz], [rEl]])))[0][0]

	# Build the output array and loop over all azimuths and elevations
	output = numpy.zeros((360,90))
	for az in list(range(360)):
		rAz = az*numpy.pi/180.0
		for el in list(range(90)):
			rEl = el*numpy.pi/180.0

			# Display the progress meter if the `progress' keyword is set to True.  The
			# progress meter displays a `.' every 2% complete and the percentages every
			# 10%.  At 100%, `Done' is displayed.
			if progress:
				fracDone = (az*90+el) / 32400.0 * 100
				if fracDone % 10 == 0 and round(fracDone,2) != 100:
					sys.stdout.write("%i%%" % fracDone)
				elif round(fracDone,2) == 100:
					sys.stdout.write("Done\n")
				elif round(fracDone,3) % 2 == 0:
					sys.stdout.write(".")
				else:
					pass
				sys.stdout.flush()

			if usePool:
				task = taskPool.apply_async(__intBeepAndSweep, args=(stands, arrayXYZ, t, az, el), kwds={'dlyCache': dlyCache, 'beamShape': beamShape[az,el], 'sampleRate': sampleRate, 'direction': (azimuth, elevation)})
				taskList.append((az,el,task))
			else:
				# Unit vector for the currect on-sky location
				currPos = numpy.array([numpy.cos(rEl)*numpy.sin(rAz), 
								numpy.cos(rEl)*numpy.cos(rAz), 
								numpy.sin(rEl)])
				# Stand response in this direction
				currResponse = standBeam.response(aipy.coord.azalt2top(numpy.concatenate([[rAz], [rEl]])))[0][0]

				# Loop over stands to build the simulated singnals
				signals = numpy.zeros((len(stands), 1000))
				for i in list(range(len(stands))):
					currDelay = dlyCache.cableDelay(stands[i]) - numpy.dot(currPos, arrayXYZ[i,:]) / c
					signals[i,:] = currResponse * numpy.cos(2*numpy.pi*49.0e6*(t + currDelay))

				# Beamform with delay-and-sum and store the RMS result
				beam = intDelayAndSum(stands, signals, sampleRate=sampleRate, azimuth=azimuth, elevation=elevation)
				output[az,el] = numpy.sqrt((beam**2).mean())

	# If pooling... Close the pool so that it knows that no ones else is joining.
	# Then, join the workers together and wait on the last one to finish before
	# saving the results.
	if usePool:
		taskPool.close()
		taskPool.join()

		# This is where he taskList list comes in handy.  We now know who did what
		# when we unpack the various results
		for az,el,task in taskList:
			output[az,el] = task.get()

		# Destroy the taskPool
		del(taskPool)

	# Done
	return output


def fftDelayAndSum(stands, data, sampleRate=dp_common.fS, LFFT=256, CentralFreq=49.0e6, azimuth=0.0, elevation=90.0):
	"""Given a list of stands and a data stream of the form stands x times, 
	delay and sum the data stream into one beam.  The delays first applied 
	as integer sample delays.  Then, the data are transformed to the 
	frequency domain and the remainder of the delay is applied as a phase 
	rotation.  A tuple consisting of the frequency array and a numpy array 
	of the frequency-domain data over time is returned."""

	# Get the stand delays in seconds
	delays = calcDelay(stands, azimuth=azimuth, elevation=elevation)

	# Make the delays into something meaningful for the shifting of the data 
	# streams.  Then, get the integer delay and sub-sample delay for each stand
	delays = delays.max() - delays
	intDelays = numpy.round(delays*sampleRate).astype(numpy.int16)
	frcDelays = delays - intDelays/sampleRate

	# Compute the frequencies of the FFT
	freq = numpy.fft.fftfreq(LFFT, d=1.0/sampleRate)
	if data.dtype.kind == 'c':
		freq = freq + CentralFreq
		freq = numpy.fft.fftshift(freq)

	# Loop over stands and FFT sections to compute the formed beam
	output = numpy.zeros((LFFT, (data.shape[1]-intDelays.max())/LFFT), dtype=numpy.complex64)
	for s in list(range(len(stands))):
		for l in list(range(output.shape[1])):
			section = data[s,(l*LFFT+intDelays[s]):((l+1)*LFFT+intDelays[s])]
			sectionF = numpy.fft.fft(section)
			if data.dtype.kind == 'c':
				sectionF = numpy.fft.fftshift(sectionF)
			sectionF *= numpy.exp(2j*numpy.pi*freq*frcDelays[s])
			output[:,l] = output[:,l] + sectionF

	# Done
	return (freq, output)


def fftBeamShape(stands, sampleRate=dp_common.fS, LFFT=256, CentralFreq=49.0e6, azimuth=0.0, elevation=90.0, progress=False):
	"""Given a list of stands, compute the on-sky response of the delay-and-sum
	scheme implemented in intDelayAndSum.  A 360x90 numpy array spaning azimuth
	and elevation is returned."""

	# Get the stand delays and convert the delay times from seconds to samples
	delays = calcDelay(stands, freq=CentralFreq, azimuth=azimuth, elevation=elevation)
	delays = numpy.round(delays*sampleRate).astype(numpy.int16)

	# Build up a base time array, load in the cable delays, and get the stand 
	# positions for geometric delay calculations.
	t = numpy.arange(0,1000)/sampleRate
	dlyCache = uvUtils.CableCache(CentralFreq, applyDispersion=True)
	xyz = uvUtils.getXYZ(stands)
	arrayX = xyz[:,0].mean()
	arrayY = xyz[:,1].mean()
	arrayZ = xyz[:,2].mean()
	arrayXYZ = xyz - numpy.array([arrayX, arrayY, arrayZ])

	# Load in the respoonse of a single isolated stand
	standBeam = __loadStandResponse(freq=CentralFreq)

	# Build the output array and loop over all azimuths and elevations
	output = numpy.zeros((360,90))
	for az in list(range(360)):
		rAz = az*numpy.pi/180.0
		for el in list(range(90)):
			rEl = el*numpy.pi/180.0

			# Display the progress meter if the `progress' keyword is set to True.  The
			# progress meter displays a `.' every 2% complete and the percentages every
			# 10%.  At 100%, `Done' is displayed.
			if progress:
				fracDone = (az*90+el) / 32400.0 * 100
				if fracDone % 10 == 0 and round(fracDone,2) != 100:
					sys.stdout.write("%i%%" % fracDone)
				elif round(fracDone,2) == 100:
					sys.stdout.write("Done\n")
				elif round(fracDone,3) % 2 == 0:
					sys.stdout.write(".")
				else:
					pass
				sys.stdout.flush()

			# Unit vector for the currect on-sky location
			currPos = numpy.array([numpy.cos(rEl)*numpy.sin(rAz), 
							numpy.cos(rEl)*numpy.cos(rAz), 
							numpy.sin(rEl)])
			# Stand response in this direction
			currResponse = standBeam.response(aipy.coord.azalt2top(numpy.concatenate([[rAz], [rEl]])))[0][0]

			# Loop over stands to build the simulated singnals
			signals = numpy.zeros((len(stands), 1000))
			for i in list(range(len(stands))):
				currDelay = dlyCache.cableDelay(stands[i]) - numpy.dot(currPos, arrayXYZ[i,:]) / c
				signals[i,:] = currResponse * numpy.cos(2*numpy.pi*CentralFreq*(t + currDelay))

			# Beamform with delay-and-sum and store the RMS result
			freq, beam = fftDelayAndSum(stands, signals, sampleRate=sampleRate, LFFT=LFFT, CentralFreq=CentralFreq, 
								azimuth=azimuth, elevation=elevation)
			output[az,el] = (numpy.abs(beam)**2).max()

	# Done
	return output


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Example script to read in the positions of stands at LWA-1 and make a plot
of the site."""

import os
import sys
import math
import numpy

import lsl.correlator.uvUtils as uvUtils
import lsl.common.stations as lwa_common

import matplotlib.pyplot as plt
from matplotlib.ticker import NullFormatter

def main(args):
	# Set the LWA Station
	station = lwa_common.lwa1()

	# Load in the stand position data
	data = uvUtils.getXYZ(numpy.arange(1,257))

	# Color-code the stands by their elevation
	color = data[:,2]

	# Plot the stands as colored circles
	fig = plt.figure(figsize=(8,8))

	ax1 = plt.axes([0.30, 0.30, 0.60, 0.60])
	ax2 = plt.axes([0.30, 0.05, 0.60, 0.15])
	ax3 = plt.axes([0.05, 0.30, 0.15, 0.60])
	ax4 = plt.axes([0.05, 0.05, 0.15, 0.15])
	c = ax1.scatter(data[:,0], data[:,1], c=color, s=40.0, alpha=0.50)
	ax1.set_xlabel('$\Delta$X [E-W; m]')
	ax1.set_xlim([-80, 80])
	ax1.set_ylabel('$\Delta$Y [N-S; m]')
	ax1.set_ylim([-80, 80])
	ax1.set_title('%s Site:  %.3f$^\circ$N, %.3f$^\circ$W' % (station.name, station.lat*180.0/math.pi, -station.long*180.0/math.pi))

	ax2.scatter(data[:,0], data[:,2], c=color, s=40.0)
	ax2.xaxis.set_major_formatter( NullFormatter() )
	ax2.set_ylabel('$\Delta$Z [m]')
	ax3.scatter(data[:,2], data[:,1], c=color, s=40.0)
	ax3.yaxis.set_major_formatter( NullFormatter() )
	ax3.set_xlabel('$\Delta$Z [m]')

	# Add and elevation colorbar to the right-hand side of the figure
	cb = plt.colorbar(c, cax=ax4, orientation='vertical', ticks=[-2, -1, 0, 1, 2])

	# Set the axis limits
	ax1.set_xlim([-60, 60])
	ax1.set_ylim([-60, 60])
	ax2.set_xlim( ax1.get_xlim() )
	ax3.set_ylim( ax1.get_ylim() )

	plt.show()
	fig.savefig('stands.png')


if __name__ == "__main__":
	main(sys.argv[1:])


# -*- coding: utf-8 -*-

"""Module that contains common values found DP ICD, revision I.  The values 
are:
  * f_S - Sampleng rate in samples per second
  * T - Slot duration in seconds
  * T_2 - Sub-slot duration
  * N_MAX_UDP - Maximum UDP packet size
"""

__version__ = '0.2'
__revision__ = '$ Revision: 5 $'
__all__ = ['fS', 'T', 'T2', 'N_MAX', '__version__', '__revision__', '__all__']

fS = 196.0e6	# Hz
T = 1.0		# seconds
T2 = 0.010	# seconds
N_MAX = 8192	# bytes






#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Example script for reading in TBN data and writing it to a TS-FITS file."""

import os
import sys
import time
from lsl.reader import tbn
from lsl.reader import errors
from lsl.writer import tsfits

import matplotlib.pyplot as plt

def main(args):
	nSamples = os.path.getsize(args[0]) / tbn.FrameSize
	print "Samples in file: ", nSamples
	fh = open(args[0], "rb", buffering=tbn.FrameSize)
	
	test = tbn.readFrame(fh)
	print "TBN Data:  %s" % test.header.isTBN()
	if not test.header.isTBN():
		raise notTBNError()
	fh.seek(0)

	nFpO = tbn.getFramesPerObs(fh)
	print "Samples per observations: %i in x pol., %i in y pol." % nFpO
	nFpO = nFpO[0] + nFpO[1]

	sampleRate = tbn.getSampleRate(fh, nFrames=nFpO)
	print "Filter code is: %i" % tbn.getSampleRate(fh, nFrames=nFpO, FilterCode=True)
	print "Sampling rate is: %i Hz" % sampleRate

	tStart = time.time()

	# Create a new FITS file with the name 'tbw.fits'
	fitsFile = tsfits.TBN('tbn-tsfits-test.fits')

	nSamples = 340000

	blocks = []
	count = {}
	syncCount = 0
	masterCount = 0
	for i in range(nSamples):
		try:
			cFrame = tbn.readBlock(fh, nFrames=nFpO, SampleRate=sampleRate)
		except errors.eofError:
			break
		except errors.syncError:
			syncCount = syncCount + 1
			continue
		except errors.numpyError:
			break
		blocks.append(cFrame)

		for frame in blocks[-1].x:
			if frame is None:
				print "sync error"
				continue
			stand, pol = frame.parseID()
			if stand not in count.keys():
				count[stand] = 0
			count[stand] = count[stand] + 1

		for xFrame, yFrame in zip(cFrame.x, cFrame.y):
			if xFrame is not None:
				fitsFile.addStandData(xFrame)
			if yFrame is not None:
				fitsFile.addStandData(yFrame)

		masterCount = masterCount + 1
		#print cFrame.header.parseID(),cFrame.data.timeTag

	tEnd = time.time()
	print 'Read %i frames in %0.3f s (%0.1f frames/s)' % (nFpO*nSamples, (tEnd-tStart), nFpO*nSamples/(tEnd-tStart))

	fh.close()
	fitsFile.close()
	fitsFile.info()

	# Summary information about the file that was just read in
	print "Summary:"
	for stand in sorted(count.keys()):
		print "Stand: %2i, Frames: %5i" % (stand, count[stand])
	print "Sync Errors: %5i" % syncCount


if __name__ == "__main__":
	main(sys.argv[1:])



# -*- coding: utf-8 -*-

"""Unit test for the lsl.correlator.uvUtils module."""

import unittest
import numpy

from lsl.correlator import uvUtils


__revision__ = "$Revision:1 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"


class uvUtils_tests(unittest.TestCase):
	"""A unittest.TestCase collection of unit tests for the lsl.correlator.uvUtils
	module."""

	def test_position(self):
		"""Test positions directly and via the position cache object."""

		# One stand
		xyz = uvUtils.getXYZ(1)
		self.assertAlmostEqual(xyz[0,0], -0.67)
		self.assertAlmostEqual(xyz[0,1], -54.63)
		self.assertAlmostEqual(xyz[0,2], 1.63)

		cache = uvUtils.PositionCache()
		xyz = cache.getXYZ(1)
		self.assertAlmostEqual(xyz[0,0], -0.67)
		self.assertAlmostEqual(xyz[0,1], -54.63)
		self.assertAlmostEqual(xyz[0,2], 1.63)

	def test_relative(self):
		"""Test baseline positions directly and via the position cache object."""

		x, y, z = uvUtils.getRelativeXYZ(1, 1)
		self.assertAlmostEqual(x, 0)
		self.assertAlmostEqual(y, 0)
		self.assertAlmostEqual(z, 0)

		x, y, z = uvUtils.getRelativeXYZ(1, 2)
		self.assertAlmostEqual(x, 1.21)
		self.assertAlmostEqual(y, 5.30)
		self.assertAlmostEqual(z, -0.13)

		cache = uvUtils.PositionCache()
		x, y, z = cache.getRelativeXYZ(1, 1)
		self.assertAlmostEqual(x, 0)
		self.assertAlmostEqual(y, 0)
		self.assertAlmostEqual(z, 0)

		x, y, z = cache.getRelativeXYZ(1, 2)
		self.assertAlmostEqual(x, 1.21)
		self.assertAlmostEqual(y, 5.30)
		self.assertAlmostEqual(z, -0.13)

	def test_position_bounds(self):
		"""Test that stand ID number bounds are respected by the position routines."""

		self.assertRaises(uvUtils.uvUtilsError, uvUtils.getXYZ, -1)
		self.assertRaises(uvUtils.uvUtilsError, uvUtils.getXYZ, 0)
		self.assertRaises(uvUtils.uvUtilsError, uvUtils.getXYZ, 259)
		self.assertRaises(uvUtils.uvUtilsError, uvUtils.getXYZ, 500)

		cache = uvUtils.PositionCache()
		self.assertRaises(uvUtils.uvUtilsError, cache.getXYZ, -1)
		self.assertRaises(uvUtils.uvUtilsError, cache.getXYZ, 0)
		self.assertRaises(uvUtils.uvUtilsError, cache.getXYZ, 259)
		self.assertRaises(uvUtils.uvUtilsError, cache.getXYZ, 500)

	def test_cable(self):
		"""Test cable delays directly and via the cable cache object."""

		ACCURACY = 12

		d = uvUtils.cableDelay(1, 10e6)
		self.assertAlmostEqual(d, 557.913e-9, ACCURACY)

		cache = uvUtils.CableCache(10e6)
		d = cache.cableDelay(1)
		self.assertAlmostEqual(d, 557.913e-9, ACCURACY)

	def test_cable_bounds(self):
		"""Test that stand ID number bounds are respected by the cable routines."""

		self.assertRaises(uvUtils.uvUtilsError, uvUtils.cableDelay, -1, 10e6)
		self.assertRaises(uvUtils.uvUtilsError, uvUtils.cableDelay, 0, 10e6)
		self.assertRaises(uvUtils.uvUtilsError, uvUtils.cableDelay, 259, 10e6)
		self.assertRaises(uvUtils.uvUtilsError, uvUtils.cableDelay, 500, 10e6)

		cache = uvUtils.CableCache(10e6)
		self.assertRaises(uvUtils.uvUtilsError, cache.cableDelay, -1)
		self.assertRaises(uvUtils.uvUtilsError, cache.cableDelay, 0)
		self.assertRaises(uvUtils.uvUtilsError, cache.cableDelay, 259)
		self.assertRaises(uvUtils.uvUtilsError, cache.cableDelay, 500)

	def test_attenuation(self):
		"""Test that the cable attenuation is working."""

		ACCURACY = 4

		a = uvUtils.cableAttenuation(1)
		self.assertAlmostEqual(a, 4.45246, ACCURACY)

	def test_attenuation_bounds(self):
		"""Test that stand ID number bound are respected by the attenuation routine."""

		self.assertRaises(uvUtils.uvUtilsError, uvUtils.cableAttenuation, -1)
		self.assertRaises(uvUtils.uvUtilsError, uvUtils.cableAttenuation, 0)
		self.assertRaises(uvUtils.uvUtilsError, uvUtils.cableAttenuation, 259)
		self.assertRaises(uvUtils.uvUtilsError, uvUtils.cableAttenuation, 500)

	def test_baseline_gen(self):
		"""Test that the generated baselines contain the correct numbers of elements."""

		standList = numpy.array([100, 101, 102, 103])

		bl = uvUtils.getBaselines(standList, IncludeAuto=False, Indicies=False)
		self.assertEqual(len(bl), 6)
		bl = uvUtils.getBaselines(standList, IncludeAuto=True, Indicies=False)
		self.assertEqual(len(bl), 10)

	def test_baseline_ind(self):
		"""Test that the baselines generated with Indicies do return indicies and vice
		versa."""

		standList = numpy.array([100, 101, 102, 103])

		bl = uvUtils.getBaselines(standList, IncludeAuto=False, Indicies=False)
		bl = numpy.array(bl)
		self.assertTrue(bl.min() == 100)
		bl = uvUtils.getBaselines(standList, IncludeAuto=True, Indicies=False)
		bl = numpy.array(bl)
		self.assertTrue(bl.min() == 100)

		bl = uvUtils.getBaselines(standList, IncludeAuto=False, Indicies=True)
		bl = numpy.array(bl)
		self.assertTrue(bl.max() < 100)
		bl = uvUtils.getBaselines(standList, IncludeAuto=True, Indicies=True)
		bl = numpy.array(bl)
		self.assertTrue(bl.max() < 100)
		
		
class uvUtils_test_suite(unittest.TestSuite):
	"""A unittest.TestSuite class which contains all of the lsl.reader units 
	tests."""
	
	def __init__(self):
		unittest.TestSuite.__init__(self)
		
		loader = unittest.TestLoader()
		self.addTests(loader.loadTestsFromTestCase(uvUtils_tests)) 


if __name__ == '__main__':
	unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Example script for reading in TBW data and writing it to a TS-FITS file."""

import os
import sys
import time
from lsl.reader import tbw
from lsl.reader import errors
from lsl.writer import tsfits

import matplotlib.pyplot as plt

def main(args):
	# Determine the number of samples in the specified file
	nSamples = os.path.getsize(args[0]) / tbw.FrameSize
	print "Samples in file: ", nSamples
	fh = open(args[0], "rb", buffering=tbw.FrameSize)

	# Make sure that the data is TBW and determine the data length
	test = tbw.readFrame(fh)
	print "TBW Data:  %s" % test.header.isTBW()
	if not test.header.isTBW():
		raise errors.notTBWError()
	print "Data Length: %i bits" % test.getDataBits()
	if test.header.getDataBits() == 12:
		nData = 400
	else:
		nData = 1200
	fh.seek(0)

	# Due to the size of the FITS files being generated, the number of frames that 
	# can be read in is limited to 300,000, or 30,000 frames for 10 stands.  Getting
	# around this limit will require finding out how to do on-the-fly FITS binary 
	# table resizing.  
	nSamples = 900000

	tStart = time.time()

	# Create a new FITS file with the name 'tbw.fits'
	fitsFile = tsfits.TBW('tbw-tsfits-test.fits')

	# Read in the data and add it to the FITS file created above
	count = {}
	syncCount = 0
	masterCount = 0
	for i in range(nSamples):
		# Read in the next frame and anticipate any problems that could occur
		try:
			cFrame = tbw.readFrame(fh, Verbose=False)
		except errors.eofError:
			break
		except errors.syncError:
			#print "WARNING: Mark 5C sync error on frame #%i" % (int(fh.tell())/TBWFrameSize-1)
			syncCount = syncCount + 1
			continue
		except errors.numpyError:
			break

		stand = cFrame.header.parseID()
		if cFrame.header.frameCount % 10000 == 0:
			print "%2i  %14i  %6.3f  %5i  %5i" % (stand, cFrame.data.timeTag, cFrame.getTime(), cFrame.header.frameCount, cFrame.header.secondsCount)
		if stand not in count.keys():
			count[stand] = 0

		fitsFile.addStandData(cFrame)

		count[stand] = count[stand] + 1
		masterCount = masterCount + 1

	tEnd = time.time()
	print 'Read %i frames in %0.3f s (%0.1f frames/s)' % (masterCount, (tEnd-tStart), masterCount/(tEnd-tStart))

	fh.close()
	fitsFile.close()
	fitsFile.info()

	# Summary information about the file that was just read in
	print "Summary:"
	for stand in sorted(count.keys()):
		print "Stand: %2i, Frames: %5i" % (stand, count[stand])
	print "Sync Errors: %5i" % syncCount


if __name__ == "__main__":
	main(sys.argv[1:])
